import panel as pn
import param

CSS = """
body {
    overflow-y: hidden;
}

.pnx-nav-buttons {
    display: block;
    text-align: left;
    text-transform: uppercase;
    width: 100%;

}
.pnx-nav-buttons details {
    outline:none;
    width: 100%;
}
.pnx-nav-buttons summary {
    outline:none;
    width: 100%;
}

.pnx-nav-buttons a, .pnx-nav-buttons summary {
    display: block;
    text-decoration: none;
    padding-top: 0.5em;
    padding-bottom: 0.5em;
    padding-right: 1em;
    padding-left: 1em;
}
.pnx-nav-buttons a:hover {
    border-radius: 2px;
}
.pnx-nav-buttons a:active {
    border-radius: 2px;
}
"""

LINK_HTML = """
<a href="{link}">{name}</a>
"""

GROUP_HTML = """
<details><summary>{title}</summary>
{links_str}
{sub_groups_str}
</details>
"""

MENU_HTML = """
<div class="pnx-nav-buttons">
{groups_str}
</div>

"""

def make_group(title, links):
    sub_links = []
    sub_groups = []
    for name, link in links.items():
        if isinstance(link, dict):
            sub_groups.append(make_group(name, link)) 
        else:
            sub_links.append(LINK_HTML.format(name=name, link=link))
    links_str = "\n".join(sub_links)
    sub_groups_str = "\n".join(sub_groups)
    return GROUP_HTML.format(title=title, links_str=links_str, sub_groups_str=sub_groups_str)

def make_menu_html(links):
    groups = [make_group(k, v) for k,v in links.items()]
    groups_str = "\n".join(groups)
    return MENU_HTML.format(groups_str=groups_str)

def recursive_get(tree, path, sep="."):
    if path in tree:
        return tree[path]
    key, _, subpath = path.partition(sep)
    if key in tree:
        return recursive_get(tree[key], subpath)
    return None

class EveWebClient(param.Parameterized):
    resources = param.Dict({})
    path_args = param.String("")
    location = param.Parameter()
    title = param.String("Eve Client")
    width = param.Integer(default=250)

    def __init__(self, **params):
        if "location" not in params:
            params["location"] = pn.state.location
        super().__init__(**params)

    def template(self, **kwargs):
        self.location = pn.state.location
        pn.config.raw_css.append(CSS)
        tmpl = pn.template.MaterialTemplate(**kwargs)
        tmpl.sidebar[:] = [self.menu_view()]
        tmpl.main[:] = [pn.panel("# No resource selected.")]
        self._template = tmpl
        return tmpl

    @param.depends("location.search", watch=True)
    def main_view(self):
        resource_path = pn.state.session_args.get("resource", [b""])[0].decode()
        resource = recursive_get(self.resources, resource_path)
        if resource is not None:
            self._template.main[:] = [resource.panel()]

    @param.depends("resources")
    def menu_view(self):
        return self.make_menu(self.resources, max_width=self.width)
    
    def make_paths(self, tree: dict, parent="",):
        paths = {}
        for k,v in tree.items():
            path = ".".join([parent, k]).strip(".")
            if isinstance(v, dict):
                paths[k] = self.make_paths(v, path)
            else:
                paths[k] = f"?resource={path}"
        return paths

    def make_menu(self, tree, max_width=250):
        paths = self.make_paths(tree)
        html = make_menu_html(paths)
        return pn.pane.HTML(html, max_width=max_width)

    def show(self):
        return self.template().show()


"""
Eve model
==========
Base classes for objects that represent Eve models.
"""

import panel as pn
import param
from copy import copy

from .settings import config as settings


class DefaultLayout(pn.GridBox):
    ncols = param.Integer(max(1, int(settings.GUI_WIDTH / 200)))
    width = param.Integer(settings.GUI_WIDTH)


class EveModelBase(param.Parameterized):
    _panel = param.ClassSelector(pn.viewable.Viewable,
                                 default=None,
                                 precedence=-1)
    max_width = param.Integer(default=settings.GUI_WIDTH, precedence=-1)
    max_height = param.Integer(default=settings.GUI_HEIGHT, precedence=-1)
    sizing_mode = param.Selector(default=settings.SIZING_MODE, precedence=-1, objects=["stretch_width", 
                                "stretch_height", "stretch_both", "scale_width", "scale_height", "scale_both"])

    def make_panel(self):
        parameters = [
            k for k, v in self.param.get_param_values()
             if not k.startswith("_") and not isinstance(v,bytes)]
        panel = pn.Param(self.param,
                         max_width=self.max_width,
                         max_height=self.max_height,
                         sizing_mode=self.sizing_mode,
                         parameters=parameters,
                         default_layout=pn.Card)
        return panel

    def panel(self):
        if self._panel is None:
            self._panel = self.make_panel()
        return self._panel

    def _repr_mimebundle_(self, include=None, exclude=None):
        mimebundle = self.panel()._repr_mimebundle_(include, exclude)
        return mimebundle

    @property
    def gui(self):
        return self.panel()

    def show(self):
        return self.panel().show()

    def servable(self):
        return self.panel().servable()

    def clone(self, **kwargs):
        params = {}
        for k,v in self.param.get_param_values():
            if isinstance(v, EveModelBase):
                try:
                    params[k] = v.clone()
                except:
                    params[k] = v
            else:
                try:
                    params[k] = copy(v)
                except:
                    params[k] = v
        params.update(kwargs)
        params["_panel"] = None
        return self.__class__(**params)

    def propagate(self, **kwargs):
        for k,v in kwargs.items():
            if hasattr(self, k):
                setattr(self, k, v)
        for k,v in self.param.get_param_values():
            if isinstance(v, EveModelBase):
                v.propagate(**kwargs)
        self._panel = None
        return self

    def __iter__(self):
        for k,v in self.param.get_param_values():
            if isinstance(v, EveModelBase):
                v = dict(v)
            yield k,v

    def __getstate__(self):
        state = super().__getstate__()
        state.pop("_panel", None)
        return state

    def __repr__(self):
        return self.name


import param
import httpx
import webbrowser
import time
import secrets
from contextlib import contextmanager, asynccontextmanager
import json
import panel as pn

from .eve_model import EveModelBase
from .settings import config as settings
from .auth import EveAuthBase, AUTH_CLASSES, DEFAULT_AUTH
from .exceptions import ServerError, AuthError
from .utils import is_valid_url


class EveSessionBase(EveModelBase):
    EXTRA_HEADERS = {
        "Accept": "application/json",
        # "Content-Type": "application/json",
    }

    log = param.String()

    auth_schemes = param.Dict(default={}, precedence=-1)
    auth_scheme = param.Selector(default=DEFAULT_AUTH, objects=list(AUTH_CLASSES))

    # selected_server = param.String("default")

    known_servers = param.Dict({})

    server_url = param.Selector(objects={"localhost": "http://localhost"})

    """Base class for Eve authentication scheme

    Inheritance:
        param.Parameterized:

    """

    def __init__(self, **params):
        auth_schemes = {name: klass() for name, klass in AUTH_CLASSES.items()}
        params["auth_schemes"] = params.get("auth_schemes", auth_schemes)
        super().__init__(**params)
        self.update_server_url_options()
        

    @classmethod
    def from_json(cls, obj):
        raise NotImplementedError
    
    @classmethod
    def from_file(cls, file_):
        raise NotImplementedError

    def update_server_url_options(self):
        self.param.server_url.names = self.known_servers
        self.param.server_url.objects = list(self.known_servers.values())
        
    @property
    def app(self):
        return None

    @property
    def auth(self):
        return self.auth_schemes.get(self.auth_scheme)
        
    def to_json(self):
        raise NotImplementedError

    def to_file(self):
        raise NotImplementedError

    def login(self, *args, **kwargs):
        """perform any actions required to aquire credentials.

        Returns:
            bool: whether login was successful
        """
        return self.auth.login(*args, **kwargs)

    def logout(self):
        return self.auth.logout()

    @property
    def logged_in(self):
        return self.auth.logged_in

    def set_credentials(self, *args, **kwargs):
        """Set the credentials manually.

        Args:
            token ([type]): [description]
        """
        self.auth.set_credentials(*args, **kwargs)

    def get_credentials(self):
        self.auth.get_credentials()

    def set_server_url(self, **kwargs):
        self.known_servers.update({name: url for name, url in kwargs.items() if is_valid_url(url)})
        self.update_server_url_options()

    def get_server_url(self, name):
        return self.known_servers.get(name, None)

    def select_server(self, name=None):
        self.update_server_url_options()
        try:
            while name not in self.known_servers:
                try:
                    name = list(self.known_servers)[int(name)]
                    break
                except:
                    pass
                options = "\n".join(["Known servers:"]+[f"{i}) {k} ({v})" for i, (k,v) 
                                                        in enumerate(self.known_servers.items())])
                name = input(options+"\nServer name: ")
            if name in self.known_servers:
                self.server_url = self.known_servers[name]
        except KeyboardInterrupt:
            pass
    
    @property
    def auth_options(self):
        return list(self.auth_schemes)

    def set_auth(self, auth_name):
        if auth_name in self.auth_schemes:
            self.auth_scheme = auth_name
        else:
            raise ValueError(f"{auth_name} is not a valid auth option. \n Valid options are: {self.auth_options}")

    @param.depends("auth_scheme")
    def credentials_view(self):
        return pn.Column(self.auth.credentials_view)
    
    def make_panel(self):
        log = pn.Param(self.param.log, 
            widgets={"log": {"type": pn.widgets.TextAreaInput, "disabled": True, "height": 150}},
            sizing_mode="stretch_both")
        server_select = pn.Param(self.param.server_url)
        server_selected = pn.Param(self.param.server_url, 
                            widgets={"server_url": pn.widgets.StaticText}, show_labels=False)
        return pn.Column(server_select, server_selected, 
                        self.param.auth_scheme,
                        self.credentials_view, log)

    def auth_view(self):
        return pn.Row()
    
    def log_error(self, msg):
        self.log = "\n".join([str(msg), self.log])
        lines = self.log.split("\n")
        if len(lines)> settings.MAX_LOG_SIZE:
            self.log = "\n".join(lines[:settings.MAX_LOG_SIZE])

    def clear_messages(self):
        pass

    @property
    def headers(self):
        h = dict(self.auth.get_headers())
        h.update(self.EXTRA_HEADERS)
        return h

    def check_errors(self, response):
        try:
            response.raise_for_status()
            if response.text:
                r = response.json()
                if "_status" in r and r["_status"]=="ERR":
                    raise ServerError(r["_error"]["code"], r["_error"]["message"])
        except Exception as e:
            if settings.IGNORE_ERRORS:
                self.log_error(e)
            else:
                raise e

    def response_hook(self, response):
        self.check_errors(response)
            
    async def response_hook_async(self, response):
        self.check_errors(response)
        
    def get_client_kwargs(self, headers={}, **kwargs):
        if not self.logged_in and not settings.IGNORE_ERRORS:
            raise AuthError("Not logged in.")
            # try:
            #     self.login()
            # except:
            #     if not settings.IGNORE_ERRORS:
            #         raise AuthError("Cannot log in.")
        timeout = kwargs.get("timeout", httpx.Timeout(settings.DEFAULT_TIMEOUT, 
                                        connect=settings.DEFAULT_TIMEOUT*6))
        user_headers = dict(headers)
        headers = self.headers
        headers.update(user_headers)
        kwargs["headers"] = headers
        kwargs["base_url"] = kwargs.get("base_url", self.server_url)
        kwargs["app"] = kwargs.get("app", self.app)
        kwargs["timeout"] = timeout
        return kwargs
    
    @contextmanager
    def Client(self, *args, **kwargs):
        kwargs = self.get_client_kwargs(**kwargs)
        client = httpx.Client(*args, **kwargs)
        client.event_hooks["response"] = [self.response_hook]
        try:
            yield client
        finally:
            client.close()

    @asynccontextmanager
    async def AsyncClient(self, *args, **kwargs ):
        kwargs = self.get_client_kwargs(**kwargs)
        client = httpx.AsyncClient(*args, **kwargs)
        client.event_hooks["response"] = [self.response_hook_async]
        try:
            yield client
        finally:
            await client.aclose()
    
    def get(self, url, timeout=10, **params):
      with self.Client() as client:
        resp = client.get(url,
                        params=params,
                        timeout=timeout)
        if resp.is_error:
            self.log_error(resp.text)
            return {}
        else:
            self.clear_messages()
            return resp.json()

    async def get_async(self, url, timeout=10, **params):
      with self.AsyncClient() as client:
        resp = await client.get(url,
                        params=params,
                        timeout=timeout)
        if resp.is_error:
            self.log_error(resp.text)
        else:
            return resp.json()
        return {}

    def post(self, url, data="", json={}, timeout=10, **kwargs):
        with self.Client() as client:
            resp = client.post(url,
                                data=data,
                                json=json,
                                headers=self.headers(),
                                timeout=timeout,
                                **kwargs)
            if resp.is_error:
                self.log_error(resp.text)
                return False
            else:
                self.clear_messages()
                return True

    async def post_async(self, url, data="", json={}, timeout=10, **kwargs):
        with self.AsyncClient() as client:
            resp = await client.post(url,
                                data=data,
                                json=json,
                                headers=self.headers(),
                                timeout=timeout,
                                **kwargs)
            if resp.is_error:
                self.log_error(resp.text)
                return False
            else:
                self.clear_messages()
                return True

    def put(self, url, data={}, etag=None, timeout=10, headers={}, **kwargs):
        with self.Client() as client:
            if etag:
                headers["If-Match"] = etag
            resp = client.put(url,
                                data=data,
                                headers=headers,
                                timeout=timeout,
                                **kwargs)
            if resp.is_error:
                self.log_error(resp.text)
                return False
            else:
                self.clear_messages()
                return True


    def patch(self, url, data, etag=None, timeout=10, headers={}, **kwargs):
        with self.Client() as client:
            if etag:
                headers["If-Match"] = etag
            try:
                resp = client.patch(url,
                                    data=data,
                                    headers=headers,
                                    timeout=timeout,
                                    **kwargs)
                if resp.is_error or settings.DEBUG:
                    self.log_error(resp.text)
                    return False
                else:
                    self.clear_messages()
                    return True
            except Exception as e:
                self.log_error(e)

    def delete(self, url, etag="", timeout=10, headers={}):
        with self.Client() as client:
            if etag:
                headers["If-Match"] = etag
            try:
                resp = client.delete(url, headers=headers, timeout=timeout)
                if resp.is_error:
                    self.log_error(resp.text)
                    return False
                else:
                    self.clear_messages()
                    return True
            except Exception as e:
                self.log_error(e)


class EveSession(EveSessionBase):
    pass

class EveSelfServeSession(EveSessionBase):
    app_settings = param.Dict(default=None, allow_None=True)
    _app = None

    @property
    def app(self):
        import eve
        if self._app is None:
            self._app = eve.Eve(settings=self.app_settings)
        return self._app


DEFAULT_SESSION_CLASS = EveSession




import base64
import panel as pn
import param
import getpass
import logging
from .base import EveAuthBase


logger = logging.getLogger(__name__)


class EveBasicAuth(EveAuthBase):
    """
    Support for eve basic auth.

    Inheritance:
        EveAuthBase:

    """
    
    username = param.String(precedence=1, doc="Basic auth username")
    password = param.String(precedence=2, doc="Basic auth password")

    @property
    def token(self):
        return base64.b64encode(":".join([self.username, self.password]).encode())

    def login(self, username=None, password=None):
        if username is None:
            username = input("Username: ")
        self.username = username
        if password is None:
            password = getpass.getpass(prompt='Password: ', stream=None) 
        self.password = password
        return True
    
    @property
    def logged_in(self):
        return self.username and self.password

    def logout(self):
        self.username = ""
        self.password = ""
        
    def get_headers(self):
        return {"Authorization": f"Basic {self.token}"}

    def make_panel(self):
        return pn.Param(self.param,
                        max_width=self.max_width,
                        max_height=self.max_height,
                        sizing_mode=self.sizing_mode,
                        parameters=["username", "password"],
                        widgets={"password": pn.widgets.PasswordInput})

    def credentials_view(self):
        return pn.Param(self.param,
                        parameters=["username", "password"],
                        widgets={"password": pn.widgets.PasswordInput},
                        max_width=self.max_width,
                        max_height=self.max_height,
                        sizing_mode=self.sizing_mode,
                        default_layout=pn.Row)



import panel as pn
import param

css = '''
.bk.panel-menu-button {
  background: #f0f0f0;
  border-radius: 0px;
  border: 1px black solid;
}
'''

class Menu(param.Parameterized):
    resources = param.Dict({})
    selected = param.String()
    resource_by_name = param.Dict({})
    width = param.Integer(default=250)
         
    @param.depends("selected")
    def selected_view(self):
        if self.selected and self.selected in self.resource_by_name:
            item = self.resource_by_name[self.selected]
            return item.panel()
        else:
            return pn.Column("# No resource selected.")
        
    @param.depends("resources")
    def menu_view(self):
        menu = self.generate_menu(self.resources, width=self.width)
        view = pn.Column(
            pn.pane.Markdown("## Resources", margin=2),
            pn.layout.Divider(margin=2,),
            menu,
            pn.layout.Divider(margin=2,),
            width_policy='max',
            sizing_mode='stretch_width',
            max_width=self.width,
        )
        return view
    
    def make_callback(self, name):
        def cb(event):
            self.selected = name
        cb.__name__ = f"{name}_callback"
        return cb
    
    def generate_menu(self, tree, width=250):
        groups = pn.layout.Accordion(width=width, toggle=True, margin=(5, 5, 5, 5))
        buttons = []
        for k,v in tree.items():
            if isinstance(v, dict):
                groups.append((k.replace("_", " ").title(), self.generate_menu(v, width=width-10)))
            else:
                self.resource_by_name[v.name] = v
                button = pn.widgets.Button(name=f"{k.replace('_', ' ').title()}",
                    max_height=70,
                    sizing_mode="stretch_width",
                    max_width=width,
                    css_classes=['panel-menu-button'],
                    margin=(0,0,0,0), 
                    )
                button.on_click(self.make_callback(v.name))
                buttons.append(button)
            
        if len(groups):
            items = [groups] + buttons
        else:
            items = buttons
        return pn.Column(
                *items, 
                width_policy='max',
                sizing_mode='stretch_width',
                max_width=width)


# -*- coding: utf-8 -*-
"""Eve-panel.

A marriage between Eve and Panel.

Example:
    
Todo:


.. _Google Python Style Guide:
   http://google.github.io/styleguide/pyguide.html

"""

__author__ = """Yossi Mosbacher"""
__email__ = 'joe.mosbacher@gmail.com'
__version__ = '0.3.13'

import panel as pn

from .auth import EveAuthBase
from .eve_client import EveClient
from .item import EveItem
from .resource import EveResource
# from .utils import from_app_config
from .menu import Menu, css
from .web_client import EveWebClient


def extension():
   pn.extension('ace')
   pn.config.raw_css.append(css)
   try:
      import holoviews as hv
      hv.extension('bokeh')
   except ImportError:
      print("Cannot import holoviews, plotting will not work.")

notebook = output_notebook = extension


"""
Eve client
====================================
Client for single or multiple Eve APIs.
"""

from pprint import pprint

import eve
import panel as pn
import param
from collections import defaultdict

from .eve_model import EveModelBase
from .session import DEFAULT_SESSION_CLASS, EveSessionBase
from .settings import config as settings
from .resource import EveResource


class EveClient(EveModelBase):
    name = param.String("EveClient", doc="Human readable name of the client")
    session = param.ClassSelector(EveSessionBase, constant=True, precedence=-1)

    @classmethod
    def from_domain_def(cls,
                        domain_def,
                        name="EveApp",
                        session=None,
                        auth_scheme=None,
                        servers={},
                        sort_by_url=False):
    
        if session is None:
            session = DEFAULT_SESSION_CLASS(known_servers=servers, auth_scheme=auth_scheme)

        if sort_by_url:
            domain_def = {v["url"]: v for v in domain_def.values()}

        sub_domains = defaultdict(dict)
        params = {}
        kwargs = {}
        for url, resource_def in sorted(domain_def.items(),
                                        key=lambda x: len(x[0])):
            sub_url, _, rest = url.partition("/")
            if rest:
                sub_domains[sub_url][rest] = resource_def
            else:
                resource = EveResource.from_resource_def(
                    resource_def, url, session=session)
                params[sub_url] = param.ClassSelector(resource.__class__,
                                                      default=resource,
                                                      constant=True)
                kwargs[sub_url] = resource
        for url, domain_def in sub_domains.items():
            if url in params:
                for sub_url, resource_def in domain_def.items():
                    resource = EveResource.from_resource_def(
                        resource_def, url, session=session)
                    kwargs[url + "_" + sub_url] = resource
                    params[url + "_" + sub_url] = param.ClassSelector(
                        resource.__class__, default=resource, constant=True)
            else:
                sub_domain = EveClient.from_domain_def(domain_def,
                                                       url,
                                                       session=session)
                kwargs[url] = sub_domain
                params[url] = param.ClassSelector(EveClient,
                                                  default=sub_domain,
                                                  constant=True)

        klass = type(name.title() + "Client", (cls, ), params)
        instance = klass(name=name, session=session, **kwargs)
        return instance

    @classmethod
    def from_app(cls, app, **kwargs):
        domain_def = app.config["DOMAIN"]
        instance = cls.from_domain_def(domain_def=domain_def, **kwargs)
        return instance
        
    @property
    def logged_in(self):
        return self.session.logged_in
    
    def login(self, *args, **kwargs):
        return self.session.login(*args, **kwargs)

    def logout(self):
        return self.session.logout()

    @property
    def resources(self):
        return {k:v for k, v in self.param.get_param_values() if isinstance(v, EveResource)}
    
    @property
    def sub_resources(self):
        return {k:v for k, v in self.param.get_param_values() if isinstance(v, EveClient)}

    @property
    def resource_tree(self):
        return self.collect_resource_tree(False)

    def make_panel(self, show_client=True, tabs_location='above'):
        tabs = [
            (k.upper().replace("_", " "),
             getattr(self, k).make_panel(show_client=False,
                                         tabs_location="above"))
            for k, v in self.param.objects().items()
            if isinstance(v, param.ClassSelector) and v.class_ in (EveClient,
                                                                   EveResource)
        ]
        tabs.sort(key=lambda x: len(x[0]))
        if show_client:
            tabs = [("Session", self.session.panel())] + tabs
        view = pn.Tabs(*tabs,
                    width=self.max_width-10,
                    max_width=self.max_width,
                    height=self.max_height,
                    sizing_mode=self.sizing_mode,
                    width_policy="max",
                    dynamic=True,
                    tabs_location=tabs_location)
        return view

    def set_token(self, token):
        self.session.set_credentials(token=token)
    
    def set_auth(self, name):
        self.session.set_auth(name)

    def set_credentials(self, **credentials):
        self.session.set_credentials(**credentials)

    def set_server_url(self, **kwargs):
        self.session.set_server_url(**kwargs)

    def get_server_url(self, name):
        self.session.get_server_url(name)

    def select_server(self, name=None):
        self.session.select_server(name)

    @property    
    def servers(self):
        return self.session.known_servers

    def collect_resource_tree(self, sort=True):
        tree = {}
        for k, v in self.param.get_param_values():
            if isinstance(v, EveClient):
                tree[k] = v.collect_resource_tree()
            elif isinstance(v, EveResource):
                tree[k] = v
        if sort:
            tree = dict(sorted(tree.items(), key=lambda x: len(x[0])))
        return tree

    def __dir__(self):
        return list(self.params())

    def __repr__(self):
        return f"EveClient(name={self.name})"
    



import logging
import param
import panel as pn

from ..eve_model import EveModelBase
from ..settings import config as settings


logger = logging.getLogger(__name__)


class EveAuthBase(EveModelBase):
    """Base class for Eve authentication scheme

    Inheritance:
        param.Parameterized:

    """

    post_login_hooks = param.HookList([])

    def get_headers(self):
        """Generate auth headers for HTTP requests.

        Returns:
            dict: Auth related headers to be included in all requests.
        """
        raise NotImplementedError

    def login(self):
        """perform any actions required to aquire credentials.

        Returns:
            bool: whether login was successful
        """
        raise NotImplementedError

    def logout(self):
        """perform any actions required to logout.

        Returns:
            bool: whether login was successful
        """
        raise NotImplementedError

    @property
    def logged_in(self):
        raise NotImplementedError

    def set_credentials(self, **credentials):
        """Set the access credentials manually.
        """
        for k,v in credentials.items():
            setattr(self, k, v)
    
    def credentials_view(self):
        raise NotImplementedError

    def post_login(self):
        for hook in self.post_login_hooks:
            hook(self)

class EveNoAuth(EveAuthBase):
    def get_headers(self):
        """Generate auth headers for HTTP requests.

        Returns:
            dict: Auth related headers to be included in all requests.
        """
        return {}

    def login(self):
        """perform any actions required to aquire credentials.

        Returns:
            bool: whether login was successful
        """
        return True

    def logout(self):
        """perform any actions required to logout.

        Returns:
            bool: whether login was successful
        """
        return True

    def credentials_view(self):
        return pn.Row()

    @property
    def logged_in(self):
        return True
